from typing import *
import textwrap
import re
import copy
import json
import sys
import copy
from collections import defaultdict

from agtest.ast import *


def eprint(*args: Any, **kwargs: Any) -> None:
    print(*args, file=sys.stderr, **kwargs)


global i
i = 0


def gensym(prefix: str = "", suffix: str = "") -> str:
    global i
    presan = re.sub("[^0-9a-zA-Z]+", "_", prefix)
    sufsan = re.sub("[^0-9a-zA-Z]+", "_", suffix)
    i += 1
    return f"{presan}{i}{sufsan}"


class NodeDesc:
    def __init__(self, node: Node):
        self.node = node
        self.name = node.name
        self.nonterminal = gensym(node.name.lower())


class ParseEquation:
    def __init__(self, name: str, syms: List[str], pyty: str, assigns: List[str], pycode: str):
        self.name = name
        self.syms = syms
        self.pyty = pyty
        self.assigns = assigns
        self.pycode = pycode


class GenResult:
    def __init__(self, program: List[Decl]):
        self.program = program
        self.literals: Dict[str, str] = dict()
        self.parse_rules: defaultdict[str, List[ParseEquation]] = defaultdict(list)
        self.starts: Set[str] = set()
        self.extra: str = ""
        self.nonterminal_map: Dict[str, str] = dict()

        self.ifaces: Dict[str, Iface] = dict()
        self.what_ifaces: Dict[str, Set[str]] = dict()
        self.what_fields: Dict[str, Dict[str, str]] = dict()
        self.node_map: Dict[str, NodeDesc] = dict()

        self.builtins: Dict[str, str] = {
            "parseInt": "",
        }

    @property
    def trans_def(self) -> str:
        s = []
        for name, rules in self.parse_rules.items():
            possible_returns = ", ".join(map(lambda e: e.pyty, rules))

            n = name.lstrip("?")
            code = textwrap.dedent(
                f"""
                def {n}(self, items: List[Union[{possible_returns}]]) -> {n}:
                    return items[0]
            """
            )
            s.append(code)

            for equation in rules:
                assigns = []
                for name, value in equation.assigns:
                    assigns.append(f"res.{name} = {value}")
                assigns_str = "\n".join(assigns)
                assigns_str = textwrap.indent(assigns_str, "    ")

                pycode = textwrap.indent(equation.pycode, "    ")
                code = textwrap.dedent(
                    """
                    def {equation.name}(self, items: Any) -> {equation.pyty}:
                        res = {equation.pyty}()
                    {pycode}
                    {assigns_str}
                        return res
                    """
                )
                code = code.format(equation=equation, assigns_str=assigns_str, pycode=pycode)
                s.append(code)
        if not s:
            s = ["pass"]
        return textwrap.indent("\n".join(s), "    ")

    @property
    def parser_data(self) -> str:
        s = []
        for sym, pat in self.literals.items():
            s.append(f"{sym}: {pat}")
        for name, rules in self.parse_rules.items():
            names = []
            for rule in rules:
                names.append(rule.name)
                s.append(f"{rule.name}: {' '.join(rule.syms)}")
            s.append(f"{name}: {' | '.join(names)}")
        s.append("%import common.WS")
        s.append("%ignore WS")
        return "\n".join(s)

    def _collect_ifaces(self) -> None:
        """collect a list of name -> iface declarations"""
        self.ifaces = dict(
            map(
                lambda c: (c.name, cast(Iface, c)),
                filter(lambda c: isinstance(c, Iface), self.program),
            )
        )

    def _create_iface_mappings(self) -> None:
        """list of node -> iface mappings"""
        self.what_ifaces = dict()
        self.what_fields = dict()
        for node in filter(lambda c: isinstance(c, Node), self.program):
            node = cast(Node, node)
            # all_fields = dict()
            self.what_ifaces[node.name] = set(node.ifaces)
            this_fields = dict()
            for iface in node.ifaces:
                fields = self.ifaces[iface].fields
                for field in fields:
                    if field.name in this_fields:
                        raise Exception("duplicate field name")
                    this_fields[field.name] = field.ty
            self.what_fields[node.name] = this_fields

    def _collect_required_thunks(
        self, env: List[Tuple[str, NodeRef]], expr: Expr
    ) -> Dict[str, str]:
        names = dict(env)
        if isinstance(expr, ExprDot):
            return self._collect_required_thunks(env, expr.left)
        elif isinstance(expr, ExprMul):
            a = self._collect_required_thunks(env, expr.left)
            b = self._collect_required_thunks(env, expr.right)
            a.update(b)
            return a
        elif isinstance(expr, ExprAdd):
            a = self._collect_required_thunks(env, expr.left)
            b = self._collect_required_thunks(env, expr.right)
            a.update(b)
            return a
        elif isinstance(expr, ExprCall):
            return self._collect_required_thunks(env, expr.func)
        elif isinstance(expr, ExprName):
            if expr.name not in names and expr.name not in self.builtins:
                raise Exception(f"unbound name '{expr.name}'")
            return dict()
        raise Exception(f"unhandled {expr.__class__}")

    def _resolve_production(self, sym: Sym) -> Tuple[bool, str]:
        """resolving a production just means checking to make sure it's a type that exists or it's a regex"""
        if isinstance(sym, SymRename):
            if isinstance(sym.ty, NodeRefByName):
                if sym.ty.name in self.node_map:
                    return True, self.node_map[sym.ty.name].nonterminal
                else:
                    raise Exception(f"unresolved name {sym.ty.name} in production")
            elif isinstance(sym.ty, NodeRegex):
                sym_name = gensym("sym")
                self.literals[sym_name] = f"/{sym.ty.pat.pattern}/"
                return True, sym_name
        elif isinstance(sym, SymLit):
            sym_name = gensym("lit")
            # hack to make repr have double quotes
            self.literals[sym_name] = json.dumps(sym.lit)
            return False, sym_name
        raise Exception(f"unhandled {sym.__class__}")

    def _build_node_map(self) -> None:
        for _node in filter(lambda c: isinstance(c, Node), self.program):
            nd = NodeDesc(cast(Node, _node))
            self.node_map[_node.name] = nd
            self.nonterminal_map[nd.name] = nd.nonterminal

    def build(self) -> None:
        def v(name: str) -> str:
            return f"__ag_{name}"

        self._collect_ifaces()
        self._create_iface_mappings()
        self._build_node_map()

        eprint("IFACE MAPS", self.what_fields, self.what_ifaces)

        # a high-level dictionary of productions; this has sub-productions that
        # should be further expanded at a later step before converting into lark
        # code
        productions_hi: Dict[str, Union[str, List[str]]] = dict()

        for node_desc in self.node_map.values():
            assert isinstance(node_desc, NodeDesc)
            self.starts.add(node_desc.nonterminal)

            class_fields = []
            for field_name, field_ty in self.what_fields[node_desc.name].items():
                class_fields.append(f"{field_name}: Thunk[{field_ty}]")
            class_fields_str = textwrap.indent("\n".join(class_fields), "    ")

            class_decl = textwrap.dedent(
                """
                class {nonterminal}:
                {g}
                    pass
                """
            ).format(nonterminal=node_desc.nonterminal, g=class_fields_str)
            self.extra += class_decl

            # print(node_desc.name, node_desc.node.ifaces)

            for variant in node_desc.node.variants:
                v_class_name = gensym(f"{node_desc.nonterminal}_var")

                prod_name = gensym(node_desc.nonterminal + "_")

                # figure out which of the symbols are inputs (aka contain data)
                # vs. literals which only parse
                # TODO: probably should also just make literals return their values
                seq = []
                inputs = []
                renamed = dict()
                for i, sym in enumerate(variant.prod):
                    isInput, n = self._resolve_production(sym)
                    var_name = f"inp_{n}_idx{i}"

                    if isInput:
                        inputs.append((i, n, var_name))
                    seq.append(n)

                    if isinstance(sym, SymRename):
                        renamed[sym.name] = i

                for field in class_fields:
                    eprint("FIELD", field)

                # generate assignments for codegen
                input_fields = copy.deepcopy(class_fields)
                assigns = []
                for i, inp, var_name in inputs:
                    name = gensym(f"{v_class_name}_inp")
                    input_fields.append(f"{var_name}: Any = 0")
                    assigns.append((var_name, f"items[{i}]"))
                class_fields_str = textwrap.indent("\n".join(input_fields), "    ")

                class_decl = textwrap.dedent(
                    """
                    class {v_class_name}({nonterminal}):
                    {g}
                        pass
                    """
                ).format(
                    v_class_name=v_class_name, nonterminal=node_desc.nonterminal, g=class_fields_str
                )
                self.extra += class_decl

                pycode = ""
                self.parse_rules[node_desc.nonterminal].append(
                    ParseEquation(prod_name, seq, v_class_name, assigns, pycode)
                )

                # create an environment for checking the equations based on the
                # production
                env: List[Tuple[str, NodeRef]] = list()
                for sym in variant.prod:
                    if isinstance(sym, SymRename):
                        env.append((sym.name, sym.ty))

                # for each of the equations, find out what the equation is
                # trying to compute, and generate a thunk corresponding to that
                # value.
                for eq in variant.equations:
                    self._collect_required_thunks(copy.deepcopy(env), eq.rhs)
